package gov.usgs.earthquake.nshmp.netcdf.www;

import static com.google.common.base.CaseFormat.UPPER_CAMEL;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;

import java.io.IOException;
import java.util.Arrays;
import java.util.Optional;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;

import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.netcdf.data.NetcdfData;
import gov.usgs.earthquake.nshmp.netcdf.data.ScienceBaseMetadata;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.SwaggerUtils;
import gov.usgs.earthquake.nshmp.www.WsUtils.EnumSerializer;
import gov.usgs.earthquake.nshmp.www.WsUtils.NaNSerializer;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.swagger.v3.oas.models.OpenAPI;
import io.swagger.v3.parser.OpenAPIV3Parser;

public class NetcdfWsUtils {
  static final Gson GSON;

  private static final Logger LOGGER = Logger.getLogger(NetcdfWsUtils.class.getName());

  static {
    GSON = new GsonBuilder()
        .registerTypeAdapter(Imt.class, new EnumSerializer<Imt>())
        .registerTypeAdapter(Double.class, new NaNSerializer())
        .disableHtmlEscaping()
        .serializeNulls()
        .setPrettyPrinting()
        .create();
  }

  public static String getRequestUrl(HttpRequest<?> request) {
    var host = Optional.of(request.getHeaders().get("Host")).orElse("");
    host = host.endsWith("/") ? host.substring(0, host.length() - 1) : host;
    var scheme = request.getHeaders().get("X-Forwarded-Proto");
    var url = host + request.getUri().toString();
    url = scheme != null ? scheme + "://" + url : url;
    return url;
  }

  public static HttpResponse<String> handleError(
      Throwable e,
      String name,
      String url) {
    var msg = e.getMessage() + " (see logs)";
    var svcResponse = ResponseBody.error()
        .name(name)
        .request(url)
        .response(msg)
        .url(url)
        .build();
    var response = GSON.toJson(svcResponse);
    LOGGER.severe(name + " -\n" + response);
    e.printStackTrace();
    return HttpResponse.serverError(response);
  }

  public static OpenAPI getOpenApi(HttpRequest<?> request, NetcdfData netcdfData,
      String servicePath) throws IOException {
    var openApi = new OpenAPIV3Parser().read("META-INF/swagger/nshmp-ws-static.yml");
    var scienceBaseMetadata = netcdfData.scienceBaseMetadata();
    SwaggerUtils.addLocationBounds(openApi, netcdfData.minimumBounds(), netcdfData.maximumBounds());
    var components = openApi.getComponents();
    var schemas = components.getSchemas();
    SwaggerUtils.siteClassSchema(schemas, netcdfData.siteClasses());
    openApi.servers(null);

    openApi.getInfo().setTitle(scienceBaseMetadata.label);

    // Update description
    var description = new StringBuilder()
        .append(scienceBaseMetadata.description + "\n")
        .append(servicePatternSection(request, servicePath))
        .append(swaggerParameterSection(netcdfData))
        .append(swaggerScienceBaseSection(scienceBaseMetadata))
        .toString();
    openApi.getInfo().setDescription(description);

    return openApi;
  }

  static enum Key {
    LATITUDE,
    LONGITUDE,
    SITE_CLASS,
    IMT;

    @Override
    public String toString() {
      return UPPER_UNDERSCORE.to(UPPER_CAMEL, name());
    }
  }

  private static String servicePatternSection(HttpRequest<?> request, String servicePath)
      throws IOException {
    var url = NetcdfWsUtils.getRequestUrl(request);
    url = url.endsWith("/swagger") ? url.replace("/swagger", "") : url;

    return new StringBuilder()
        .append("## Service Call Patterns\n")
        .append(
            "### Query Pattern\n" +

                "The query based service call is in the form of:\n" +

                "```text\n" +
                url + servicePath + "?longitude={number}&latitude={number}&siteClass={string}\n" +
                "````\n" +

                "Example:\n" +
                "```text\n" +
                url + servicePath + "?longitude=-118&latitude=34&siteClass=BC\n" +
                "```\n")
        .append(
            "### Slash Pattern\n" +

                "The slash based service call is in the form of:\n" +

                "```text\n" +
                url + servicePath + "/{longitude}/{latitude}/{siteClass}\n" +
                "```\n" +

                "Example:\n" +

                "```text\n" +
                url + servicePath + "/-118/34/BC\n" +
                "```\n" +

                "> Note: To obtain data for all site classes simply " +
                "call the query or slashed based service with no site class parameter.\n")
        .toString();
  }

  private static String swaggerParameterSection(NetcdfData netcdfData) {
    return new StringBuilder()
        .append("\n## Parameters\n")
        .append(
            SwaggerUtils.locationBoundsInfo(netcdfData.minimumBounds(), netcdfData.maximumBounds(),
                Optional.of("### ")))
        .append(SwaggerUtils.siteClassInfo(netcdfData.siteClasses(), Optional.of("### ")))
        .toString();
  }

  private static String swaggerScienceBaseSection(ScienceBaseMetadata scienceBaseMetadata) {
    return new StringBuilder()
        .append("## ScienceBase\n")
        .append("Data history: " + scienceBaseMetadata.history)
        .append("<br><br>")
        .append(
            String.format(
                "Returned data are spatially interpolated from " +
                    "the %s grid data from ScienceBase:\n",
                scienceBaseMetadata.gridStep))
        .append(Arrays.stream(scienceBaseMetadata.scienceBaseInfo)
            .map(info -> "- " + info.url)
            .collect(Collectors.joining("\n")))
        .toString();
  }
}
